from dataclasses import dataclass, field
from typing import List, Union

from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config

from ..models import Model, ModelConfig


@dataclass
class RandomForestConfig(ModelConfig):
    _fit_input_feature_types: List[Union[None, type]] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Union[None, type]] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_process_desc: str = field(
        default="Fitting the random forest model", init=False, repr=False
    )
    predict_proba_process_desc: str = field(
        default=None,
        init=False,
        repr=False,
    )
    processor_name: str = field(default="random_forest", init=False, repr=False)

    n_estimators: int = 100
    max_depth: int = None
    min_samples_split: int = 2
    min_samples_leaf: int = 1
    min_weight_fraction_leaf: float = 0.0
    max_leaf_nodes: int = None
    min_impurity_decrease: float = 0.0
    bootstrap: bool = False
    oob_score: bool = False
    n_jobs: int = None
    random_state: int = 42
    verbose: int = 0
    warm_start: bool = False
    ccp_alpha: float = 0.0
    max_samples: int = None
    monotonic_cst: dict = None
    criterion: str = None  # gini for classification, squared_error for regression
    max_features: Union[str, int, float] = "sqrt"

    class_weight: Union[str, dict, List[dict]] = None
    n_classes: int = None

    use_predict_proba: bool = False
    task: str = None

    estimator: Union[RandomForestClassifier, RandomForestRegressor] = field(
        default=None, init=False, repr=False
    )


class RandomForestModel(Model):
    config_class = RandomForestConfig
    config: RandomForestConfig
    random_forest: Union[RandomForestClassifier, RandomForestRegressor]

    def __init__(
        self,
        n_estimators: int = 100,
        max_depth: int = None,
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        min_weight_fraction_leaf: float = 0.0,
        max_leaf_nodes: int = None,
        min_impurity_decrease: float = 0.0,
        bootstrap: bool = True,
        oob_score: bool = False,
        n_jobs: int = None,
        random_state: int = None,
        verbose: int = 0,
        warm_start: bool = False,
        ccp_alpha: float = 0.0,
        max_samples: int = None,
        monotonic_cst: dict = None,
        criterion: str = None,
        max_features: Union[str, int, float] = "sqrt",
        class_weight: Union[str, dict, List[dict]] = None,
        task: str = None,
        config: RandomForestConfig = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
            monotonic_cst=monotonic_cst,
            criterion=criterion,
            max_features=max_features,
            task=task,
            class_weight=class_weight,
            **kwargs,
        )

        if self.config.task == "classification":
            self.random_forest = RandomForestClassifier(
                n_estimators=self.config.n_estimators,
                criterion=self.config.criterion if self.config.criterion else "gini",
                max_depth=self.config.max_depth if self.config.max_depth else None,
                min_samples_split=self.config.min_samples_split,
                min_samples_leaf=self.config.min_samples_leaf,
                min_weight_fraction_leaf=self.config.min_weight_fraction_leaf,
                max_features=self.config.max_features,
                max_leaf_nodes=self.config.max_leaf_nodes,
                min_impurity_decrease=self.config.min_impurity_decrease,
                bootstrap=self.config.bootstrap,
                oob_score=self.config.oob_score,
                n_jobs=self.config.n_jobs,
                random_state=self.config.random_state,
                verbose=self.config.verbose,
                warm_start=self.config.warm_start,
                class_weight=self.config.class_weight
                if self.config.class_weight and self.config.class_weight != "None"
                else None,
                ccp_alpha=self.config.ccp_alpha,
                max_samples=self.config.max_samples,
                monotonic_cst=self.config.monotonic_cst,
            )
        else:
            self.random_forest = RandomForestRegressor(
                n_estimators=self.config.n_estimators,
                criterion=self.config.criterion
                if self.config.criterion
                else "squared_error",
                max_depth=self.config.max_depth if self.config.max_depth > 0 else None,
                min_samples_split=self.config.min_samples_split,
                min_samples_leaf=self.config.min_samples_leaf,
                min_weight_fraction_leaf=self.config.min_weight_fraction_leaf,
                max_features=self.config.max_features,
                max_leaf_nodes=self.config.max_leaf_nodes,
                min_impurity_decrease=self.config.min_impurity_decrease,
                bootstrap=self.config.bootstrap,
                oob_score=self.config.oob_score,
                n_jobs=self.config.n_jobs,
                random_state=self.config.random_state,
                verbose=self.config.verbose,
                warm_start=self.config.warm_start,
                ccp_alpha=self.config.ccp_alpha,
                max_samples=self.config.max_samples,
                monotonic_cst=self.config.monotonic_cst,
            )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.task == "classification":
            self.random_forest = RandomForestClassifier(
                n_estimators=self.config.n_estimators,
                criterion=self.config.criterion if self.config.criterion else "gini",
                max_depth=self.config.max_depth if self.config.max_depth else None,
                min_samples_split=self.config.min_samples_split,
                min_samples_leaf=self.config.min_samples_leaf,
                min_weight_fraction_leaf=self.config.min_weight_fraction_leaf,
                max_features=self.config.max_features,
                max_leaf_nodes=self.config.max_leaf_nodes,
                min_impurity_decrease=self.config.min_impurity_decrease,
                bootstrap=self.config.bootstrap,
                oob_score=self.config.oob_score,
                n_jobs=self.config.n_jobs,
                random_state=self.config.random_state,
                verbose=self.config.verbose,
                warm_start=self.config.warm_start,
                class_weight=self.config.class_weight
                if self.config.class_weight and self.config.class_weight != "None"
                else None,
                ccp_alpha=self.config.ccp_alpha,
                max_samples=self.config.max_samples,
                monotonic_cst=self.config.monotonic_cst,
            )
        else:
            self.random_forest = RandomForestRegressor(
                n_estimators=self.config.n_estimators,
                criterion=self.config.criterion
                if self.config.criterion
                else "squared_error",
                max_depth=self.config.max_depth if self.config.max_depth > 0 else None,
                min_samples_split=self.config.min_samples_split,
                min_samples_leaf=self.config.min_samples_leaf,
                min_weight_fraction_leaf=self.config.min_weight_fraction_leaf,
                max_features=self.config.max_features,
                max_leaf_nodes=self.config.max_leaf_nodes,
                min_impurity_decrease=self.config.min_impurity_decrease,
                bootstrap=self.config.bootstrap,
                oob_score=self.config.oob_score,
                n_jobs=self.config.n_jobs,
                random_state=self.config.random_state,
                verbose=self.config.verbose,
                warm_start=self.config.warm_start,
                ccp_alpha=self.config.ccp_alpha,
                max_samples=self.config.max_samples,
                monotonic_cst=self.config.monotonic_cst,
            )
        if self.config.use_predict_proba:
            self.output_dtype = "float32"
        return self

    @property
    def feature_importances_(self):
        return self.config.estimator.feature_importances_

    @property
    def feature_names_in_(self):
        return self.config.feature_names_in_

    def fit(
        self,
        X,
        y=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "RandomForestModel":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, target_column
        )
        return self._process_fit(
            X,
            y,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def predict(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._method_prefix = "_predict"
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def predict_proba(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._method_prefix = "_predict_proba"
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_sklearn(self, X, y):
        self.config.estimator = self.random_forest.fit(X, y)
        return self

    def _predict_sklearn(self, X):
        return self.config.estimator.predict(X)

    def _predict_proba_sklearn(self, X):
        return self.config.estimator.predict_proba(X)

    @staticmethod
    def suggest_params(data):
        params = {
            "n_estimators": ("suggest_int", (100, 2000), {}),
            "max_depth": ("suggest_int", (0, 11), {}),
            "min_samples_split": ("suggest_int", (2, 20), {}),
            "min_samples_leaf": ("suggest_int", (1, 20), {}),
            # "bootstrap": ("suggest_categorical", ([True, False],), {}),
            # "max_features": ("suggest_categorical", (["sqrt", "log2"],), {}),
            "n_jobs": -1,
        }

        return params

    @staticmethod
    def suggest_first_trial():
        return {
            "n_estimators": 100,
            "max_depth": 0,
            "min_samples_split": 2,
            "min_samples_leaf": 1,
            "n_jobs": -1,
        }


class RandomForestForClassification(RandomForestModel):
    """Random Forest model for classification tasks"""

    @property
    def classes_(self):
        return self.config.estimator.classes_

    @sync_backup_config
    def set_params(self, **kwargs):
        kwargs.pop("task", None)
        self.config.task = "classification"
        return super().set_params(**kwargs)

    @staticmethod
    def suggest_params(data):
        params = RandomForestModel.suggest_params(data)
        params["criterion"] = ("suggest_categorical", (["gini", "entropy"],), {})
        params["class_weight"] = (
            "suggest_categorical",
            (["balanced", "balanced_subsample", "None"],),
            {},
        )
        return params

    @staticmethod
    def suggest_first_trial():
        return {
            **RandomForestModel.suggest_first_trial(),
            "criterion": "gini",
            "class_weight": "balanced",
        }


class RandomForestForRegression(RandomForestModel):
    @sync_backup_config
    def set_params(self, **kwargs):
        kwargs.pop("task", None)
        self.config.task = "regression"
        return super().set_params(**kwargs)

    @staticmethod
    def suggest_params(data):
        params = RandomForestModel.suggest_params(data)
        params["criterion"] = (
            "suggest_categorical",
            (["squared_error", "absolute_error", "poisson"],),
            {},
        )
        return params



# Copyright 2024 Patrick Smyth and Hugging Face authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import os
import queue
import random
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from queue import Empty
from typing import Callable, Iterable, Set, TypeVar, Union

import multiprocess
import multiprocess.pool
import numpy as np
from biocore.utils.import_util import (
    is_datasets_available,
    is_polars_available,
    is_tf_available,
    is_torch_available,
)

Y = TypeVar("Y")


def is_temporal(val):
    return isinstance(val, (datetime, date, time, timedelta))


def is_decimal(val):
    return isinstance(val, Decimal)


def as_py(val):
    # return as int64
    if isinstance(val, datetime):
        return val.timestamp()
    elif isinstance(val, date):
        return val.toordinal()
    elif isinstance(val, time):
        return val.hour * 3600 + val.minute * 60 + val.second + val.microsecond / 1e6
    elif isinstance(val, timedelta):
        return val.total_seconds()
    elif isinstance(val, Decimal):
        return float(val)
    elif isinstance(val, np.float16):
        return float(val)
    elif isinstance(val, bytes):
        return f"base64:{val.hex()}"
    elif isinstance(val, dict):
        return {k: as_py(v) for k, v in val.items()}
    return val


def enable_full_determinism(seed: int, warn_only: bool = False):
    """
    Helper function for reproducible behavior during distributed training. See
    - https://pytorch.org/docs/stable/notes/randomness.html for pytorch
    - https://www.tensorflow.org/api_docs/python/tf/config/experimental/enable_op_determinism for tensorflow
    """
    # set seed first
    set_seed(seed)

    if is_torch_available():
        import torch

        # Enable PyTorch deterministic mode. This potentially requires either the environment
        # variable 'CUDA_LAUNCH_BLOCKING' or 'CUBLAS_WORKSPACE_CONFIG' to be set,
        # depending on the CUDA version, so we set them both here
        os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
        torch.use_deterministic_algorithms(True, warn_only=warn_only)

        # Enable CUDNN deterministic mode
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    if is_tf_available():
        import tensorflow as tf

        tf.config.experimental.enable_op_determinism()


def set_seed(seed: int):
    """
    Helper function for reproducible behavior to set the seed in `random`, `numpy`, `torch` and/or `tf` (if installed).

    Args:
        seed (`int`): The seed to set.
    """
    random.seed(seed)
    np.random.seed(seed)

    if is_polars_available():
        from polars import set_random_seed

        set_random_seed(seed)

    def is_torch_npu_available():
        try:
            import torch

            return torch.npu.is_available()
        except ImportError:
            return False

    if is_torch_available():
        import torch

        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    if is_tf_available():
        import tensorflow as tf

        tf.random.set_seed(seed)


# This function is a copy of the one in datasets.utils.py_util.py, Copyright 2024
# Hugging Face authors. Licensed under the Apache 2.0 license. See the license file for
# details at https://www.apache.org/licenses/LICENSE-2.0
def _get_pool_pid(
    pool: Union[multiprocessing.pool.Pool, multiprocess.pool.Pool],
) -> Set[int]:
    return {f.pid for f in pool._pool}


# This function is a copy of the one in datasets.utils.py_util.py, Copyright 2024
# Hugging Face authors. Licensed under the Apache 2.0 license. See the license file for
# details at https://www.apache.org/licenses/LICENSE-2.0
def _write_generator_to_queue(
    queue: queue.Queue, func: Callable[..., Iterable[Y]], kwargs: dict
) -> int:
    for i, result in enumerate(func(**kwargs)):
        queue.put(result)
    return i


# This function is a copy of the one in datasets.utils.py_util.py, Copyright 2024
# Hugging Face authors. Licensed under the Apache 2.0 license. See the license file for
# details at https://www.apache.org/licenses/LICENSE-2.0
def iflatmap_unordered(
    pool: Union[multiprocessing.pool.Pool, multiprocess.pool.Pool],
    func: Callable[..., Iterable[Y]],
    *,
    kwargs_iterable: Iterable[dict],
) -> Iterable[Y]:
    """
    Similar to `itertools.chain.from_iterable(map(func, iterable))` but with a potentially more efficient implementation
    that doesn't require storing all the intermediate results in memory.

    Args:
        func (Callable): The function to apply to each element of the input iterable.
        iterable (Iterable): The input iterable.

    Returns:
        Iterator: The flattened iterable.
    """
    if is_datasets_available():
        from datasets.utils.py_utils import iflatmap_unordered

        return iflatmap_unordered(pool, func, kwargs_iterable=kwargs_iterable)
    else:
        initial_pool_pid = _get_pool_pid(pool)
        pool_changed = False
        manager_cls = (
            multiprocessing.Manager
            if isinstance(pool, multiprocessing.pool.Pool)
            else multiprocess.Manager
        )
        with manager_cls() as manager:
            queue = manager.Queue()
            async_results = [
                pool.apply_async(_write_generator_to_queue, (queue, func, kwargs))
                for kwargs in kwargs_iterable
            ]
            try:
                while True:
                    try:
                        yield queue.get(timeout=0.05)
                    except Empty:
                        if (
                            all(async_result.ready() for async_result in async_results)
                            and queue.empty()
                        ):
                            break
                    if _get_pool_pid(pool) != initial_pool_pid:
                        pool_changed = True
                        # One of the subprocesses has died. We should not wait forever.
                        raise RuntimeError(
                            "One of the subprocesses has abruptly died during map operation."
                            "To debug the error, disable multiprocessing."
                        )
            finally:
                if not pool_changed:
                    # we get the result in case there's an error to raise
                    [async_result.get(timeout=0.05) for async_result in async_results]



from dataclasses import dataclass, field
from typing import List, Optional, Union

from sklearn.linear_model import LogisticRegression

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..models import Model, ModelConfig

logger = logging.get_logger(__name__)


@dataclass
class LogisticRegressionConfig(ModelConfig):
    _fit_input_feature_types: List[Union[None, type]] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Union[None, type]] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _fit_process_desc: str = field(
        default="Fitting the lasso model", init=False, repr=False
    )
    predict_process_desc: str = field(
        default=None,
        init=False,
        repr=False,
    )
    predict_proba_process_desc: str = field(
        default=None,
        init=False,
        repr=False,
    )
    processor_name: str = field(default="logistic_regression", init=False, repr=False)

    input_columns: SelectedColumnTypes = None
    target_column: SelectedColumnTypes = None

    penalty: str = "l2"
    dual: bool = False
    tol: float = 1e-4
    C: float = 1.0
    fit_intercept: bool = True
    intercept_scaling: float = 1
    class_weight: Union[dict, str] = None
    random_state: Optional[int] = 42
    solver: str = "saga"
    max_iter: int = 100
    multi_class: str = "auto"
    verbose: int = 0
    warm_start: bool = False
    n_jobs: Optional[int] = None
    l1_ratio: Optional[float] = None

    use_predict_proba: bool = False
    task: str = field(default="classification", init=False, repr=False)
    estimator: LogisticRegression = field(default=None, init=False, repr=False)
    n_classes: int = None

    def __post_init__(self):
        self._fit_process_desc = "Fitting the logistic regression model"
        self.predict_process_desc = "Predicting with the logistic regression model"
        self.predict_proba_process_desc = (
            "Predicting probabilities with the logistic regression model"
        )


class LogisticRegressionModel(Model):
    config_class = LogisticRegressionConfig
    config: LogisticRegressionConfig
    logistic_regression: LogisticRegression

    def __init__(
        self,
        penalty: str = "l2",
        dual: bool = False,
        tol: float = 1e-4,
        C: float = 1.0,
        fit_intercept: bool = True,
        intercept_scaling: float = 1,
        class_weight: Union[dict, str] = None,
        random_state: int = 42,
        solver: str = "lbfgs",
        max_iter: int = 100,
        multi_class: str = "auto",
        verbose: int = 0,
        warm_start: bool = False,
        n_jobs: Optional[int] = None,
        l1_ratio: Optional[float] = None,
        config: LogisticRegressionConfig = None,
        **kwargs,
    ):
        super().__init__(
            penalty=penalty,
            dual=dual,
            tol=tol,
            C=C,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            random_state=random_state,
            solver=solver,
            max_iter=max_iter,
            multi_class=multi_class if multi_class != "None" else None,
            verbose=verbose,
            warm_start=warm_start,
            n_jobs=n_jobs,
            l1_ratio=l1_ratio,
            config=config,
            **kwargs,
        )
        self.logistic_regression = LogisticRegression(
            penalty=self.config.penalty,
            dual=self.config.dual,
            tol=self.config.tol,
            C=self.config.C,
            fit_intercept=self.config.fit_intercept,
            intercept_scaling=self.config.intercept_scaling,
            class_weight=self.config.class_weight
            if self.config.class_weight != "None"
            else None,
            random_state=self.config.random_state,
            solver=self.config.solver,
            max_iter=self.config.max_iter,
            multi_class=self.config.multi_class,
            verbose=self.config.verbose,
            warm_start=self.config.warm_start,
            n_jobs=self.config.n_jobs,
            l1_ratio=self.config.l1_ratio,
        )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.logistic_regression = LogisticRegression(
            penalty=self.config.penalty,
            dual=self.config.dual,
            tol=self.config.tol,
            C=self.config.C,
            fit_intercept=self.config.fit_intercept,
            intercept_scaling=self.config.intercept_scaling,
            class_weight=self.config.class_weight
            if self.config.class_weight != "None"
            else None,
            random_state=self.config.random_state,
            solver=self.config.solver,
            max_iter=self.config.max_iter,
            multi_class=self.config.multi_class,
            verbose=self.config.verbose,
            warm_start=self.config.warm_start,
            n_jobs=self.config.n_jobs,
            l1_ratio=self.config.l1_ratio,
        )
        return self

    @property
    def feature_importances_(self):
        return self.config.estimator.coef_.flatten()

    @property
    def feature_names_in_(self):
        return self.config.feature_names_in_

    def fit(
        self,
        X,
        y=None,
        input_columns: SelectedColumnTypes = None,
        target_column: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "LogisticRegressionModel":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns, target_column
        )
        return self._process_fit(
            X,
            y,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def _fit_sklearn(self, X, y):
        self.config.estimator = self.logistic_regression.fit(X, y)
        return self

    def _predict_sklearn(self, X):
        return self.config.estimator.predict(X)

    def _predict_proba_sklearn(self, X):
        return self.config.estimator.predict_proba(X)

    @staticmethod
    def suggest_params(data):
        params = {
            "C": (
                "suggest_float",
                (
                    1e-4,
                    1e2,
                ),
                {"log": True},
            ),
            "tol": (
                "suggest_float",
                (
                    1e-5,
                    1e-2,
                ),
                {"log": True},
            ),
            "max_iter": ("suggest_categorical", ([100, 200, 500, 1000],), {}),
            "penalty": "l1",
            "n_jobs": -1,
            "class_weight": ("suggest_categorical", (["balanced", "None"],), {}),
        }
        return params



# ruff: noqa
# from .auto import *
from .feature_selection import *
from .filtering import *
from .scaling import *
from .feature_extraction import *
from .imputation import *
from .encoding import *
from .resampling import *
from .transformation import *



from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Type

import pyarrow as pa

from biofit.integration.biosets import get_feature
from biofit.integration.R import RCaller
from biofit.integration.R.r_caller import PackageNotInstalledError
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..scaling import Scaler, ScalerConfig

logger = logging.get_logger(__name__)


@dataclass
class TMMScalerConfig(ScalerConfig):
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    # process descriptions
    _fit_process_desc: str = field(
        default="Determining reference samples for TMM normalization",
        init=False,
        repr=False,
    )
    _transform_process_desc: str = field(
        default="Applying TMM normalization", init=False, repr=False
    )
    processor_name: str = field(default="tmm", init=False, repr=False)

    # attributes
    r_source: str = field(
        default=Path(__file__).with_suffix(".R").as_posix(), init=False, repr=False
    )
    fit_func_name: str = field(default="edger_tmm_fit", init=False, repr=False)
    cpm_transform_func_name: str = field(
        default="edger_tmm_cpm_transform", init=False, repr=False
    )
    tpm_transform_func_name: str = field(
        default="edger_tmm_tpm_transform", init=False, repr=False
    )

    log: bool = True
    prior_count: int = 2
    meta_col: str = "Length"
    gene_col: str = "genes"

    # estimated attributes
    ref_samples: Optional[pa.Table] = field(default=None, init=False, repr=False)


class TMMScalerConfigForMetagenomics(TMMScalerConfig):
    # dataset specific attributes
    _input_feature_types: List[Type] = (
        get_feature("Abundance"),
        get_feature("ReadCount"),
    )
    dataset_name = "metagenomics"


class TMMScalerConfigForOTU(TMMScalerConfig):
    # dataset specific attributes
    _input_feature_types: List[Type] = get_feature("Abundance")
    dataset_name = "otu"


class TMMScalerConfigForSNP(TMMScalerConfig):
    # dataset specific attributes
    _input_feature_types: List[Type] = get_feature("GenomicVariant")
    dataset_name = "snp"


class TMMScaler(Scaler):
    output_dtype = "float64"

    # config class
    config_class = TMMScalerConfig
    config: TMMScalerConfig

    def __init__(
        self,
        log: bool = True,
        prior_count: int = 2,
        meta_col: str = "Length",
        gene_col: str = "genes",
        config: Optional[TMMScalerConfig] = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            log=log,
            prior_count=prior_count,
            meta_col=meta_col,
            gene_col=gene_col,
            **kwargs,
        )
        r_caller = RCaller.from_script(self.config.r_source)
        install_missing = kwargs.get("install_missing")
        try:
            r_caller.verify_r_dependencies(
                cran_dependencies=["BiocManager"],
                bioconductor_dependencies=["edgeR"],
                install_missing=install_missing,
            )
        except PackageNotInstalledError:
            raise PackageNotInstalledError(
                "TMMScale requires the following R package: edgeR. To "
                "install, initialize the TMMScaler with install_missing=True or run "
                "R -e 'BiocManager::install(\"edgeR\")' in your terminal."
            )
        self.fit_func = r_caller.get_method(self.config.fit_func_name)
        self.cpm_transform = r_caller.get_method(self.config.cpm_transform_func_name)
        self.tpm_transform = r_caller.get_method(self.config.tpm_transform_func_name)

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        genes=None,
        input_columns: SelectedColumnTypes = None,
        gene_col: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns, gene_col)
        return self._process_transform(
            X,
            genes,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        genes=None,
        input_columns: SelectedColumnTypes = None,
        gene_col: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            genes=genes,
            gene_col=gene_col,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)

        # load file from src/biofit/preprocessing/scaling/tmm/tmm.R
        r_caller = RCaller.from_script(self.config.r_source)
        self.fit_func = r_caller.get_method(self.config.fit_func_name)
        self.cpm_transform = r_caller.get_method(self.config.cpm_transform_func_name)
        self.tpm_transform = r_caller.get_method(self.config.tpm_transform_func_name)
        return self

    def _fit_arrow(self, X: pa.Table):
        self.config.ref_samples = self.fit_func(X)
        return self

    def _transform_arrow(self, X: pa.Table, genes=None):
        if genes:
            return self.tpm_transform(
                X=X,
                feature_meta=genes,
                ref_samples=self.config.ref_samples,
                log=self.config.log,
                prior_count=self.config.prior_count,
                meta_col=self.config.meta_col,
            )
        else:
            return self.cpm_transform(
                X=X,
                ref_samples=self.config.ref_samples,
                log=self.config.log,
                prior_count=self.config.prior_count,
            )



# ruff: noqa
from .plot_relative_abundance import (
    RelativeAbundancePlotter,
    RelativeAbundancePlotterConfig,
    RelativeAbundancePlotterConfigForASV,
    RelativeAbundancePlotterConfigForMaldi,
    RelativeAbundancePlotterConfigForMetagenomics,
    RelativeAbundancePlotterConfigForOTU,
    RelativeAbundancePlotterConfigForSNP,
)
from .relative_abundance import RelativeAbundanceScaler, RelativeAbundanceScalerConfig



# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modified by: Patrick Smyth
# Date: 2024
# Summary of changes:
# - removed `config` to track the dill version. Version is parsed directly from the dill
# package.
"""Extends `dill` to support pickling more types and produce more consistent dumps."""

import os
import sys
from io import BytesIO
from types import CodeType, FunctionType

import dill
from packaging import version


class Pickler(dill.Pickler):
    dispatch = dill._dill.MetaCatchingDict(dill.Pickler.dispatch.copy())
    _legacy_no_dict_keys_sorting = False

    def save(self, obj, save_persistent_id=True):
        obj_type = type(obj)
        if obj_type not in self.dispatch:
            if "regex" in sys.modules:
                import regex  # type: ignore

                if obj_type is regex.Pattern:
                    pklregister(obj_type)(_save_regexPattern)
            if "spacy" in sys.modules:
                import spacy  # type: ignore

                if issubclass(obj_type, spacy.Language):
                    pklregister(obj_type)(_save_spacyLanguage)
            if "tiktoken" in sys.modules:
                import tiktoken  # type: ignore

                if obj_type is tiktoken.Encoding:
                    pklregister(obj_type)(_save_tiktokenEncoding)
            if "torch" in sys.modules:
                import torch  # type: ignore

                if issubclass(obj_type, torch.Tensor):
                    pklregister(obj_type)(_save_torchTensor)

                if obj_type is torch.Generator:
                    pklregister(obj_type)(_save_torchGenerator)

                # Unwrap `torch.compile`-ed modules
                if issubclass(obj_type, torch.nn.Module):
                    obj = getattr(obj, "_orig_mod", obj)
            if "transformers" in sys.modules:
                import transformers  # type: ignore

                if issubclass(obj_type, transformers.PreTrainedTokenizerBase):
                    pklregister(obj_type)(_save_transformersPreTrainedTokenizerBase)

        # Unwrap `torch.compile`-ed functions
        if obj_type is FunctionType:
            obj = getattr(obj, "_torchdynamo_orig_callable", obj)
        dill.Pickler.save(self, obj, save_persistent_id=save_persistent_id)

    def _batch_setitems(self, items):
        if self._legacy_no_dict_keys_sorting:
            return super()._batch_setitems(items)
        # Ignore the order of keys in a dict
        try:
            # Faster, but fails for unorderable elements
            items = sorted(items)
        except Exception:  # TypeError, decimal.InvalidOperation, etc.
            from datasets.fingerprint import Hasher

            items = sorted(items, key=lambda x: Hasher.hash(x[0]))
        dill.Pickler._batch_setitems(self, items)

    def memoize(self, obj):
        # Don't memoize strings since two identical strings can have different Python ids
        if type(obj) is not str:  # noqa: E721
            dill.Pickler.memoize(self, obj)


def pklregister(t):
    """Register a custom reducer for the type."""

    def proxy(func):
        Pickler.dispatch[t] = func
        return func

    return proxy


def dump(obj, file):
    """Pickle an object to a file."""
    Pickler(file, recurse=True).dump(obj)


def dumps(obj):
    """Pickle an object to a string."""
    file = BytesIO()
    dump(obj, file)
    return file.getvalue()


if version.parse(dill.__version__) < version.parse("0.3.6"):

    def log(pickler, msg):
        dill._dill.log.info(msg)

elif version.parse(dill.__version__) in [
    version.parse("0.3.6").release,
    version.parse("0.3.7").release,
    version.parse("0.3.8").release,
]:

    def log(pickler, msg):
        dill._dill.logger.trace(pickler, msg)


@pklregister(set)
def _save_set(pickler, obj):
    log(pickler, f"Se: {obj}")
    try:
        # Faster, but fails for unorderable elements
        args = (sorted(obj),)
    except Exception:  # TypeError, decimal.InvalidOperation, etc.
        from datasets.fingerprint import Hasher

        args = (sorted(obj, key=Hasher.hash),)

    pickler.save_reduce(set, args, obj=obj)
    log(pickler, "# Se")


def _save_regexPattern(pickler, obj):
    import regex  # type: ignore

    log(pickler, f"Re: {obj}")
    args = (obj.pattern, obj.flags)
    pickler.save_reduce(regex.compile, args, obj=obj)
    log(pickler, "# Re")


def _save_tiktokenEncoding(pickler, obj):
    import tiktoken  # type: ignore

    log(pickler, f"Enc: {obj}")
    args = (obj.name, obj._pat_str, obj._mergeable_ranks, obj._special_tokens)
    pickler.save_reduce(tiktoken.Encoding, args, obj=obj)
    log(pickler, "# Enc")


def _save_torchTensor(pickler, obj):
    import torch  # type: ignore

    # `torch.from_numpy` is not picklable in `torch>=1.11.0`
    def create_torchTensor(np_array, dtype=None):
        tensor = torch.from_numpy(np_array)
        if dtype:
            tensor = tensor.type(dtype)
        return tensor

    log(pickler, f"To: {obj}")
    if obj.dtype == torch.bfloat16:
        args = (obj.detach().to(torch.float).cpu().numpy(), torch.bfloat16)
    else:
        args = (obj.detach().cpu().numpy(),)
    pickler.save_reduce(create_torchTensor, args, obj=obj)
    log(pickler, "# To")


def _save_torchGenerator(pickler, obj):
    import torch  # type: ignore

    def create_torchGenerator(state):
        generator = torch.Generator()
        generator.set_state(state)
        return generator

    log(pickler, f"Ge: {obj}")
    args = (obj.get_state(),)
    pickler.save_reduce(create_torchGenerator, args, obj=obj)
    log(pickler, "# Ge")


def _save_spacyLanguage(pickler, obj):
    import spacy  # type: ignore

    def create_spacyLanguage(config, bytes):
        lang_cls = spacy.util.get_lang_class(config["nlp"]["lang"])
        lang_inst = lang_cls.from_config(config)
        return lang_inst.from_bytes(bytes)

    log(pickler, f"Sp: {obj}")
    args = (obj.config, obj.to_bytes())
    pickler.save_reduce(create_spacyLanguage, args, obj=obj)
    log(pickler, "# Sp")


def _save_transformersPreTrainedTokenizerBase(pickler, obj):
    log(pickler, f"Tok: {obj}")
    # Ignore the `cache` attribute
    state = obj.__dict__
    if "cache" in state and isinstance(state["cache"], dict):
        state["cache"] = {}
    pickler.save_reduce(type(obj), (), state=state, obj=obj)
    log(pickler, "# Tok")


if version.parse(dill.__version__) < version.parse("0.3.6"):

    @pklregister(CodeType)
    def _save_code(pickler, obj):
        """
        From dill._dill.save_code
        This is a modified version that removes the origin (filename + line no.)
        of functions created in notebooks or shells for example.
        """
        dill._dill.log.info(f"Co: {obj}")
        # The filename of a function is the .py file where it is defined.
        # Filenames of functions created in notebooks or shells start with '<'
        # ex: <ipython-input-13-9ed2afe61d25> for ipython, and <stdin> for shell
        # Filenames of functions created in ipykernel the filename
        # look like f"{tempdir}/ipykernel_{id1}/{id2}.py"
        # Moreover lambda functions have a special name: '<lambda>'
        # ex: (lambda x: x).__code__.co_name == "<lambda>"  # True
        #
        # For the hashing mechanism we ignore where the function has been defined
        # More specifically:
        # - we ignore the filename of special functions (filename starts with '<')
        # - we always ignore the line number
        # - we only use the base name of the file instead of the whole path,
        # to be robust in case a script is moved for example.
        #
        # Only those two lines are different from the original implementation:
        co_filename = (
            ""
            if obj.co_filename.startswith("<")
            or (
                len(obj.co_filename.split(os.path.sep)) > 1
                and obj.co_filename.split(os.path.sep)[-2].startswith("ipykernel_")
            )
            or obj.co_name == "<lambda>"
            else os.path.basename(obj.co_filename)
        )
        co_firstlineno = 1
        # The rest is the same as in the original dill implementation
        if dill._dill.PY3:
            if hasattr(obj, "co_posonlyargcount"):
                args = (
                    obj.co_argcount,
                    obj.co_posonlyargcount,
                    obj.co_kwonlyargcount,
                    obj.co_nlocals,
                    obj.co_stacksize,
                    obj.co_flags,
                    obj.co_code,
                    obj.co_consts,
                    obj.co_names,
                    obj.co_varnames,
                    co_filename,
                    obj.co_name,
                    co_firstlineno,
                    obj.co_lnotab,
                    obj.co_freevars,
                    obj.co_cellvars,
                )
            else:
                args = (
                    obj.co_argcount,
                    obj.co_kwonlyargcount,
                    obj.co_nlocals,
                    obj.co_stacksize,
                    obj.co_flags,
                    obj.co_code,
                    obj.co_consts,
                    obj.co_names,
                    obj.co_varnames,
                    co_filename,
                    obj.co_name,
                    co_firstlineno,
                    obj.co_lnotab,
                    obj.co_freevars,
                    obj.co_cellvars,
                )
        else:
            args = (
                obj.co_argcount,
                obj.co_nlocals,
                obj.co_stacksize,
                obj.co_flags,
                obj.co_code,
                obj.co_consts,
                obj.co_names,
                obj.co_varnames,
                co_filename,
                obj.co_name,
                co_firstlineno,
                obj.co_lnotab,
                obj.co_freevars,
                obj.co_cellvars,
            )
        pickler.save_reduce(CodeType, args, obj=obj)
        dill._dill.log.info("# Co")
        return

elif version.parse(dill.__version__).release[:3] in [
    version.parse("0.3.6").release,
    version.parse("0.3.7").release,
    version.parse("0.3.8").release,
]:
    # From: https://github.com/uqfoundation/dill/blob/dill-0.3.6/dill/_dill.py#L1104
    @pklregister(CodeType)
    def save_code(pickler, obj):
        dill._dill.logger.trace(pickler, "Co: %s", obj)

        ############################################################################################################
        # Modification here for huggingface/datasets
        # The filename of a function is the .py file where it is defined.
        # Filenames of functions created in notebooks or shells start with '<'
        # ex: <ipython-input-13-9ed2afe61d25> for ipython, and <stdin> for shell
        # Filenames of functions created in ipykernel the filename
        # look like f"{tempdir}/ipykernel_{id1}/{id2}.py"
        # Moreover lambda functions have a special name: '<lambda>'
        # ex: (lambda x: x).__code__.co_name == "<lambda>"  # True
        #
        # For the hashing mechanism we ignore where the function has been defined
        # More specifically:
        # - we ignore the filename of special functions (filename starts with '<')
        # - we always ignore the line number
        # - we only use the base name of the file instead of the whole path,
        # to be robust in case a script is moved for example.
        #
        # Only those two lines are different from the original implementation:
        co_filename = (
            ""
            if obj.co_filename.startswith("<")
            or (
                len(obj.co_filename.split(os.path.sep)) > 1
                and obj.co_filename.split(os.path.sep)[-2].startswith("ipykernel_")
            )
            or obj.co_name == "<lambda>"
            else os.path.basename(obj.co_filename)
        )
        co_firstlineno = 1
        # The rest is the same as in the original dill implementation, except for the replacements:
        # - obj.co_filename => co_filename
        # - obj.co_firstlineno => co_firstlineno
        ############################################################################################################

        if hasattr(obj, "co_endlinetable"):  # python 3.11a (20 args)
            args = (
                obj.co_lnotab,  # for < python 3.10 [not counted in args]
                obj.co_argcount,
                obj.co_posonlyargcount,
                obj.co_kwonlyargcount,
                obj.co_nlocals,
                obj.co_stacksize,
                obj.co_flags,
                obj.co_code,
                obj.co_consts,
                obj.co_names,
                obj.co_varnames,
                co_filename,  # Modification for huggingface/datasets ############################################
                obj.co_name,
                obj.co_qualname,
                co_firstlineno,  # Modification for huggingface/datasets #########################################
                obj.co_linetable,
                obj.co_endlinetable,
                obj.co_columntable,
                obj.co_exceptiontable,
                obj.co_freevars,
                obj.co_cellvars,
            )
        elif hasattr(obj, "co_exceptiontable"):  # python 3.11 (18 args)
            args = (
                obj.co_lnotab,  # for < python 3.10 [not counted in args]
                obj.co_argcount,
                obj.co_posonlyargcount,
                obj.co_kwonlyargcount,
                obj.co_nlocals,
                obj.co_stacksize,
                obj.co_flags,
                obj.co_code,
                obj.co_consts,
                obj.co_names,
                obj.co_varnames,
                co_filename,  # Modification for huggingface/datasets ############################################
                obj.co_name,
                obj.co_qualname,
                co_firstlineno,  # Modification for huggingface/datasets #########################################
                obj.co_linetable,
                obj.co_exceptiontable,
                obj.co_freevars,
                obj.co_cellvars,
            )
        elif hasattr(obj, "co_linetable"):  # python 3.10 (16 args)
            args = (
                obj.co_lnotab,  # for < python 3.10 [not counted in args]
                obj.co_argcount,
                obj.co_posonlyargcount,
                obj.co_kwonlyargcount,
                obj.co_nlocals,
                obj.co_stacksize,
                obj.co_flags,
                obj.co_code,
                obj.co_consts,
                obj.co_names,
                obj.co_varnames,
                co_filename,  # Modification for huggingface/datasets ############################################
                obj.co_name,
                co_firstlineno,  # Modification for huggingface/datasets #########################################
                obj.co_linetable,
                obj.co_freevars,
                obj.co_cellvars,
            )
        elif hasattr(obj, "co_posonlyargcount"):  # python 3.8 (16 args)
            args = (
                obj.co_argcount,
                obj.co_posonlyargcount,
                obj.co_kwonlyargcount,
                obj.co_nlocals,
                obj.co_stacksize,
                obj.co_flags,
                obj.co_code,
                obj.co_consts,
                obj.co_names,
                obj.co_varnames,
                co_filename,  # Modification for huggingface/datasets ############################################
                obj.co_name,
                co_firstlineno,  # Modification for huggingface/datasets #########################################
                obj.co_lnotab,
                obj.co_freevars,
                obj.co_cellvars,
            )
        else:  # python 3.7 (15 args)
            args = (
                obj.co_argcount,
                obj.co_kwonlyargcount,
                obj.co_nlocals,
                obj.co_stacksize,
                obj.co_flags,
                obj.co_code,
                obj.co_consts,
                obj.co_names,
                obj.co_varnames,
                co_filename,  # Modification for huggingface/datasets ############################################
                obj.co_name,
                co_firstlineno,  # Modification for huggingface/datasets #########################################
                obj.co_lnotab,
                obj.co_freevars,
                obj.co_cellvars,
            )

        pickler.save_reduce(dill._dill._create_code, args, obj=obj)
        dill._dill.logger.trace(pickler, "# Co")
        return



# ruff: noqa
from .col_sum import *
from .col_mean import *
from .col_missingness import *
from .row_sum import *
from .row_mean import *
from .row_missingness import *
from .distance import *
from .correlation import *



from typing import List, Optional, Union, Callable
import sys
import inspect
import importlib
from functools import wraps
from pathlib import Path


from .. import config
from .file_utils import is_file_name, PathLike
from . import logging

logger = logging.get_logger(__name__)


def load_module_or_class(full_path):
    if full_path is None:
        return None
    try:
        module_path, entity_name = full_path.rsplit(".", 1)
        module = (
            importlib.import_module(module_path)
            if module_path not in sys.modules
            else sys.modules[module_path]
        )
    except ImportError:
        # Handle the case where the module can't be imported
        logger.error(f"Failed to import module: {module_path}")
        return None

    try:
        entity = (
            getattr(module, entity_name)
            if entity_name not in sys.modules
            else sys.modules[entity_name]
        )
    except AttributeError:
        # Handle the case where the entity isn't found in the module
        logger.error(f"{entity_name} not found in module: {module_path}")
        return None

    return entity


def load_method(full_path, method_name):
    entity = load_module_or_class(full_path)
    if entity:
        return getattr(entity, method_name)
    return None


def _get_cache_dir(
    cache_files: Optional[List[dict]] = None, cache_file_name: Optional[PathLike] = None
) -> Optional[Path]:
    # check if cache_file_name is a path
    cache_dir = None
    if cache_file_name is not None:
        if isinstance(cache_file_name, PathLike):
            cache_file_name = Path(cache_file_name)
            if "/" in cache_file_name.as_posix():
                cache_dir = cache_file_name.resolve().parent

    if not cache_dir and cache_files:
        cache_dir = Path(cache_files[0]["filename"]).resolve().parent

    return cache_dir


def _get_cache_info(
    cache_files: Optional[List[dict]] = None,
    cache_dir: Optional[Path] = None,
    cache_file_name: Optional[Union[str, Path]] = None,
    file_ext=".arrow",
):
    if cache_file_name:
        if is_file_name(cache_file_name):
            cache_dir = cache_dir or Path.cwd()
            cache_file_name = Path(cache_dir) / Path(cache_file_name).with_suffix(
                file_ext
            )
        else:
            cache_file_name = Path(cache_file_name).with_suffix(file_ext)

    cache_dir = _get_cache_dir(cache_files, cache_file_name)
    return cache_file_name, cache_dir


UNRECORDED_METHODS = ["train_test_split"]
_RECORDER_ACTIVE = False


def pre_recording(func, *args, **kwargs):
    new_fingerprint = kwargs.get("new_fingerprint", kwargs.get("fingerprint", None))

    signature = inspect.signature(func)
    cache_file_name = kwargs.get("cache_file_name", None)
    if not cache_file_name and "cache_file_name" in signature.parameters:
        arg_pos = list(signature.parameters).index("cache_file_name")
        if len(args) > arg_pos:
            cache_file_name = args[arg_pos]
    cache_dir = None
    self = args[0] if args else kwargs.get("self", None)
    if getattr(self, "cache_files", None):
        cache_file_name, cache_dir = _get_cache_info(
            self.cache_files, cache_dir, cache_file_name
        )
    if isinstance(cache_dir, Path):
        cache_dir = cache_dir.resolve().as_posix()
    if isinstance(cache_file_name, Path):
        cache_file_name = cache_file_name.resolve().as_posix()

    return {
        "cache_file_name": cache_file_name,
        "cache_dir": cache_dir,
        "new_fingerprint": new_fingerprint,
    }


def toggle_recorder():
    global _RECORDER_ACTIVE
    _RECORDER_ACTIVE = not _RECORDER_ACTIVE


def record_step(
    replay_func: Optional[str] = None,
    pre_recording: Optional[Callable] = pre_recording,
    post_recording: Optional[Callable] = None,
):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not config.RECORDER_ENABLED or _RECORDER_ACTIVE:
                return func(*args, **kwargs)

            toggle_recorder()
            extra_info = {}
            if pre_recording:
                extra_info = pre_recording(func, *args, **kwargs)
            out = func(*args, **kwargs)
            if post_recording:
                out = post_recording(
                    out,
                    func.__name__,
                    args,
                    kwargs,
                    replay_func=replay_func,
                    info=extra_info,
                )

            toggle_recorder()

            return out

        return wrapper

    return decorator



from dataclasses import dataclass, field
from typing import List, Type, Union

from biocore.utils.import_util import is_biosets_available
import numpy as np
import pyarrow as pa
from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging

from ..encoding import Encoder, EncoderConfig

logger = logging.get_logger(__name__)


@dataclass
class LabelBinarizerConfig(EncoderConfig):
    """Configuration class for label binarizer."""

    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ClassLabel")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("ClassLabel")], init=False, repr=False
    )

    processor_name: str = field(default="label_binarizing", init=False, repr=False)
    _transform_process_desc: str = field(default=None, init=False, repr=False)
    positive_labels: list = field(default_factory=list)
    negative_labels: list = field(default_factory=list)
    as_one_hot: bool = False
    names: list = field(default_factory=list, init=False, repr=False)

    # auto attributes
    label_mapping: dict = field(default_factory=dict, init=False, repr=False)

    def __post_init__(self):
        if self.as_one_hot:
            self._transform_process_desc = "One-hot encoding labels"
            self._fit_process_desc = "Determining unique labels"
            if self.names:
                self._transform_process_desc = (
                    f"One-hot encoding labels with {self.names}"
                )
                self.label_mapping = {
                    str(label): i for i, label in enumerate(self.names)
                }
        else:
            if not isinstance(self.positive_labels, list):
                self.positive_labels = [self.positive_labels]
            if not isinstance(self.negative_labels, list):
                self.negative_labels = [self.negative_labels]
            if len(self.positive_labels) < 1 and len(self.negative_labels) < 1:
                raise ValueError(
                    "At least one of positive_labels or negative_labels must be provided"
                )

            self._transform_process_desc = f"Binarizing labels with positive_labels={self.positive_labels} and negative_labels={self.negative_labels}"

            if self.positive_labels:
                self._transform_process_desc = (
                    f"Binarizing labels with positive_labels={self.positive_labels}"
                )

                if self.negative_labels:
                    self._transform_process_desc += (
                        f" and negative_labels={self.negative_labels}"
                    )
            else:
                self._transform_process_desc = (
                    f"Binarizing labels with negative_labels={self.negative_labels}"
                )

            if not isinstance(self.positive_labels, list):
                self.positive_labels = [self.positive_labels]

            if not isinstance(self.negative_labels, list):
                self.negative_labels = [self.negative_labels]

            self.label_mapping = {str(label): 1 for label in self.positive_labels}
            self.label_mapping.update({str(label): 0 for label in self.negative_labels})


class LabelBinarizer(Encoder):
    config_class = LabelBinarizerConfig
    config: LabelBinarizerConfig

    def __init__(
        self,
        positive_labels: list = [],
        negative_labels: list = [],
        names: list = [],
        as_one_hot: bool = False,
        config: LabelBinarizerConfig = None,
        **kwargs,
    ):
        """
        Args:
            positive_labels (list, *optional*):
                The labels to be considered as positive. Default is `[]`.
            negative_labels (list, *optional*):
                The labels to be considered as negative. Default is `[]`.
            as_one_hot (bool, *optional*):
                Whether to encode the labels as one-hot vectors. Default is `False`.
            names (list, *optional*):
                The names of the classes. This is required when `as_one_hot` is `True`.
            config (LabelBinarizerConfig, *optional*):
            **kwargs:
                Arguments that are passed to ProcessorConfig.
        """
        super().__init__(
            config=config,
            positive_labels=positive_labels,
            negative_labels=negative_labels,
            as_one_hot=as_one_hot,
            names=names,
            **kwargs,
        )
        self.output_feature_type = get_feature("BinClassLabel")

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.config.__post_init__()
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "LabelBinarizer":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _get_features_out(self, X, **kwargs):
        kwargs.pop("one_to_one_features", None)
        features = super()._get_features_out(X, one_to_one_features=True, **kwargs)

        input_names = DataHandler.get_column_names(X, generate_cols=True)
        input_names = [input_names[i] for i in self._selected_indices]
        if self.config.as_one_hot and self.config.names:
            # incase that the user provided the pre-encoded labels of get_feature("ClassLabel")
            for name in input_names:
                label_feature = features.pop(name, None)
                if isinstance(label_feature, get_feature("ClassLabel")):
                    label_names = label_feature.names
                    if self.config.names and isinstance(self.config.names[0], str):
                        self.config.label_mapping = {
                            str(label_names.index(label)): i
                            for i, label in enumerate(self.config.names)
                        }
            features.update(
                {
                    name: get_feature("ClassLabel")(names=["negative", "positive"])
                    for name in self.config.names
                }
            )
            return features

        if is_biosets_available():
            for name in input_names:
                # incase that the user provided the pre-encoded labels of get_feature("ClassLabel")
                label_feature = features.get(name)
                if isinstance(label_feature, get_feature("ClassLabel")):
                    label_names = label_feature.names
                    if self.config.positive_labels and isinstance(
                        self.config.positive_labels[0], str
                    ):
                        self.config.label_mapping = {
                            str(label_names.index(label)): 1
                            for label in self.config.positive_labels
                        }
                    if self.config.negative_labels and isinstance(
                        self.config.negative_labels[0], str
                    ):
                        self.config.label_mapping.update(
                            {
                                str(label_names.index(label)): 0
                                for label in self.config.negative_labels
                            }
                        )
                label_feature = get_feature("BinClassLabel")(
                    names=self.config.names or ["negative", "positive"],
                    positive_labels=self.config.positive_labels,
                    negative_labels=self.config.negative_labels,
                )
                features[name] = label_feature

        return features

    def _process_fit_input(self, input, **kwargs):
        if not self.config.as_one_hot or self.config.names:
            kwargs["fn_kwargs"]["fn"] = None
        if self.config.names:
            self.config.label_mapping = {
                str(label): i for i, label in enumerate(self.config.names)
            }
        return super()._process_fit_input(input, **kwargs)

    def _check_data(self, X):
        X_dims = DataHandler.get_shape(X)
        if len(X_dims) > 1:
            if X_dims[1] == 1:
                out = DataHandler.select_column(X, 0)
            else:
                raise ValueError(
                    f"Expected input to have 1 column, got {X_dims[1]} columns"
                )
        else:
            out = X
        return DataHandler.to_format(out, "list")

    def _fit_array(self, X):
        out = self._check_data(X)
        self.config.names = DataHandler.to_format(DataHandler.unique(out), "list")
        self.config._transform_process_desc = (
            f"Encoding labels with: {self.config.label_mapping}"
        )
        return self

    def _partial_fit_array(self, X):
        out = self._check_data(X)
        labs = DataHandler.to_format(DataHandler.unique(out), "list")
        self.config.names = np.unique(self.config.names + labs).tolist()
        return self

    def _pool_fit(self, fitted_processors):
        names = []
        for processor in fitted_processors:
            if isinstance(processor, LabelBinarizer):
                names += processor.config.names
        self.config.names = np.unique(names).tolist()
        self.config.label_mapping = {
            str(label): i for i, label in enumerate(self.config.names)
        }
        self.config._transform_process_desc = (
            f"Encoding labels with: {self.config.label_mapping}"
        )
        return self

    def _process_fit_output(self, input, out):
        if self.config.as_one_hot:
            self.config._n_features_out = len(self.config.names)
            self.output_dtype = "int64"
        else:
            self.config._n_features_out = None
        self.config.label_mapping = {
            str(k): v for k, v in self.config.label_mapping.items()
        }
        return super()._process_fit_output(input, out)

    def _one_hot_transform(self, X):
        col = self._check_data(X)

        labs = np.zeros((len(col), len(self.config.names) + 1), dtype=np.int64)

        inds = [self.config.label_mapping.get(str(val), -1) for val in col]
        labs[np.arange(len(col)), inds] = 1
        return labs[:, :-1]

    def _binarize_transform(self, X: Union[pa.Table, pa.Array]):
        col = self._check_data(X)

        if self.config.positive_labels and self.config.negative_labels:
            labs = [-1] * len(col)
        elif self.config.positive_labels:
            labs = [0] * len(col)
        else:
            labs = [1] * len(col)

        if self.config.positive_labels and self.config.negative_labels:
            labs = [
                self.config.label_mapping.get(str(val), -1)
                if labs[i] == -1
                else labs[i]
                for i, val in enumerate(col)
            ]
        elif self.config.positive_labels:
            labs = [
                self.config.label_mapping.get(str(val), 0) if labs[i] == 0 else labs[i]
                for i, val in enumerate(col)
            ]
        else:
            labs = [
                self.config.label_mapping.get(str(val), 1) if labs[i] == 1 else labs[i]
                for i, val in enumerate(col)
            ]

        return pa.array(labs)

    def _transform_array(self, X):
        if not self.config.as_one_hot:
            return self._binarize_transform(X)
        else:
            return self._one_hot_transform(X)



# ruff: noqa
from .file_utils import (
    add_end_docstrings,
    add_start_docstrings,
    estimate_dataset_size,
    has_ext,
    has_separator,
    hash_url_to_filename,
    is_file_name,
    is_local_path,
    is_relative_path,
    is_remote_url,
    move_temp_file,
    url_or_path_join,
    url_or_path_parent,
)
from .fingerprint import (
    Hasher,
    _build_cache_dir,
    disable_caching,
    enable_caching,
    fingerprint_from_data,
    fingerprint_from_kwargs,
    generate_cache_dir,
    get_cache_file_name,
    is_caching_enabled,
    update_fingerprint,
)
from .logging import (
    disable_progress_bar,
    enable_progress_bar,
    set_verbosity,
    set_verbosity_debug,
    set_verbosity_error,
    set_verbosity_info,
    set_verbosity_warning,
    silence,
    unsilence,
)
from .py_util import (
    as_py,
    enable_full_determinism,
    set_seed,
)
from .recorder import (
    load_module_or_class,
    record_step,
)
from .table_util import (
    concat_blocks,
    determine_upcast,
    init_arrow_buffer_and_writer,
    is_binary_like,
    is_fixed_width,
    is_large_binary_like,
    read_arrow_table,
    upcast_tables,
    write_arrow_table,
)
from .types import Unset



# ruff: noqa
from .feature_importance import (
    FeatureImportancePlotter,
    FeatureImportancePlotterConfig,
    FeatureImportancePlotterConfigForOTU,
)
from .sample_metadata import SampleMetadataPlotter, SampleMetadataPlotterConfig
from .plotting_utils import (
    generate_violin,
    generate_barplot,
    generate_scatterplot,
    generate_histogram,
    generate_comparison_histogram,
    plot_correlation,
    plot_feature_distribution,
    compare_feature_distributions,
    plot_sample_distribution,
    compare_sample_distributions,
    plot_dimension_reduction,
    plot_feature_importance,
    plot_sample_metadata,
)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Optional, Type, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..filtering import SampleFilter, SampleFilterConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class MissingLabelsSampleFilterConfig(SampleFilterConfig):
    # process description
    _transform_process_desc: str = field(
        default="SampleFilter out rows with missing labels", init=False, repr=False
    )
    processor_name: str = field(default="missing_labels", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("TARGET_FEATURE_TYPES")],
        init=False,
        repr=False,
    )

    missing_label: Optional[Union[str, int]] = "auto"

    def __post_init__(self):
        self._transform_process_desc = (
            f"SampleFiltering out rows with labels equaling to {self.missing_label}"
        )


class MissingLabelsSampleFilter(SampleFilter):
    """Remove samples that are not labeled

    - config:
    - mising_label: The value we deem as missing and want to remove. Default is "auto".
    """

    # main config class
    config_class = MissingLabelsSampleFilterConfig
    config: MissingLabelsSampleFilterConfig

    def __init__(
        self,
        config: Optional[MissingLabelsSampleFilterConfig] = None,
        missing_label: Optional[Union[str, int]] = "auto",
        **kwargs,
    ):
        super().__init__(config=config, missing_label=missing_label, **kwargs)

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "MissingLabelsSampleFilter":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _process_transform_input(self, X, **kwargs):
        selected_indices = kwargs["fn_kwargs"].get("selected_indices", None) or [0]
        if self.config.missing_label == "auto":
            if DataHandler.is_categorical(X, selected_indices[0], threshold=30):
                self.config.missing_label = -1
                kwargs["desc"] = self._transform_process_desc = (
                    "SampleFiltering out rows with labels equaling to -1"
                )
            else:
                self.config.missing_label = None
                kwargs["desc"] = self._transform_process_desc = (
                    "SampleFiltering out rows with labels equaling to None"
                )
        return super()._process_transform_input(X, **kwargs)

    def _transform_arrow(self, X: Union[pa.Table, pa.Array]):
        if isinstance(X, pa.Table):
            return (
                pc.not_equal(X.column(0), self.config.missing_label).to_numpy().tolist()
            )
        return pc.not_equal(X, self.config.missing_label).to_numpy().tolist()

    def _transform_pandas(self, X: pd.DataFrame):
        return X.ne(self.config.missing_label).values.tolist()

    def _transform_polars(self, X: "pl.DataFrame"):
        import polars as pl

        if isinstance(X, pl.Series):
            return X.ne(self.config.missing_label).to_numpy().tolist()
        return (
            X.get_column(X.columns[0]).ne(self.config.missing_label).to_numpy().tolist()
        )

    def _transform_numpy(self, X: np.ndarray):
        return X != self.config.missing_label



import argparse

from .__version__ import __version__


def main():
    print("use biofit.cli instead. available commands: --version")
    # print(f"{__file__}")
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=f"{__version__}")
    parser.parse_args()


if __name__ == "__main__":
    main()



from typing import List

from biocore.utils.import_util import is_biosets_available
from biocore.utils.inspect import get_kwargs
from sklearn.pipeline import Pipeline

from biofit.auto.auto_factory import (
    _BaseAutoProcessorClass,
    _get_class,
    _LazyAutoMapping,
)
from biofit.processing import BaseProcessor

from .configuration_auto import (
    CONFIG_MAPPING_NAMES,
    DATASET_TO_MAPPER_NAMES,
    PROCESSOR_MAPPING_NAMES,
    AutoPreprocessorConfig,
)

PROCESSOR_MAPPING = _LazyAutoMapping(CONFIG_MAPPING_NAMES, PROCESSOR_MAPPING_NAMES)


class AutoProcessor(_BaseAutoProcessorClass):
    _processor_mapping = PROCESSOR_MAPPING


class ProcessorPipeline(Pipeline):
    """A pipeline of processors."""

    def __init__(self, processors: List[BaseProcessor] = None, **kwargs):
        if "steps" in kwargs:
            return super().__init__(**kwargs)

        self.processors = processors
        steps = []
        for i, processor in enumerate(processors):
            if not isinstance(processor, tuple):
                if hasattr(processor, "config"):
                    steps.append(
                        (
                            f"{processor.config.processor_name}_{i + 1}",
                            processor,
                        )
                    )
                else:
                    steps.append((f"{processor.__class__.__name__}_{i + 1}", processor))
            else:
                steps.append(processor)
        super().__init__(steps, **kwargs)

    def fit_transform(self, X, y=None, **kwargs):
        for processor in self.processors:
            if isinstance(processor, tuple):
                p = processor[-1]
            else:
                p = processor
            fit_transform_kwargs = get_kwargs(kwargs, p.fit_transform)
            X = p.fit_transform(X, y, **fit_transform_kwargs)
        return X

    def pop(self, index):
        self.steps.pop(index)
        return self.processors.pop(index)


class AutoPreprocessor(_BaseAutoProcessorClass):
    _processor_mapping = PROCESSOR_MAPPING

    @classmethod
    def for_dataset(cls, dataset_name, **kwargs):
        """
        Create a preprocessor pipeline for a given dataset.

        Args:
            dataset: str
                The dataset name.
        Returns:
            ProcessorPipeline
                The preprocessor pipeline for the dataset.
        """
        if is_biosets_available():
            from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
        else:
            EXPERIMENT_TYPE_ALIAS = {}

        dataset_name = EXPERIMENT_TYPE_ALIAS.get(dataset_name, dataset_name)
        _processor_mapping = _LazyAutoMapping(
            DATASET_TO_MAPPER_NAMES.get(dataset_name), PROCESSOR_MAPPING_NAMES
        )
        configs = AutoPreprocessorConfig.for_dataset(dataset_name, **kwargs)
        procs = [
            _get_class(config, _processor_mapping)._from_config(config)
            for config in configs
        ]

        processors = ProcessorPipeline(procs)
        return processors



from dataclasses import dataclass, field
from typing import List, Type

import numpy as np

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..scaling import Scaler, ScalerConfig

logger = logging.get_logger(__name__)


@dataclass
class CLRScalerConfig(ScalerConfig):
    """
    Configuration for CumulativeSumScaler.
    """

    processor_name: str = field(default="css", init=False, repr=False)
    _fit_process_desc: str = field(
        default="Calculating cumulative sum scaling percentile", init=False, repr=False
    )
    _transform_process_desc: str = field(
        default="Applying cumulative sum scaling", init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )

    pseudocount: float = 1e-6
    input_columns: List[str] = None


class CLRScaler(Scaler):
    """
    BaseCumSumScaler applies cumulative sum scaling to the input data.
    """

    output_dtype = "float64"
    config_class = CLRScalerConfig
    config: CLRScalerConfig

    def __init__(
        self, pseudocount: float = 1e-6, config: CLRScalerConfig = None, **kwargs
    ):
        super().__init__(config=config, pseudocount=pseudocount, **kwargs)

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "CLRScaler":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns or self.config.input_columns
        )
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(
            input_columns or self.config.input_columns
        )
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _transform_numpy(self, X: np.ndarray):
        X = X + self.config.pseudocount
        # Calculating the geometric mean of each sample
        from scipy.stats import gmean

        r_geo_mean = gmean(X, axis=1)[:, np.newaxis]

        return np.log(X / r_geo_mean)



from dataclasses import dataclass, field

import pandas as pd
import pyarrow as pa
from biocore import DataHandler

from biofit.processing import BaseProcessor, ProcessorConfig
from biofit.utils import logging
from biofit.utils.table_util import string_to_arrow

logger = logging.get_logger(__name__)


@dataclass
class ResamplerConfig(ProcessorConfig):
    processor_type: str = field(default="resampling", init=False, repr=False)


class Resampler(BaseProcessor):
    _feature_dependent = (
        False  # SampleFiltering does not depend on features when transforming
    )

    def _process_transform_batch_output(self, input, out, **fn_kwargs):
        inds = out["indices"]
        if len(inds):
            return DataHandler.select_rows(input, inds)
        else:
            cols = DataHandler.get_column_names(input, generate_cols=True)
            schema = pa.schema(
                [
                    pa.field(k, string_to_arrow(v))
                    for k, v in DataHandler.get_dtypes(input).items()
                ]
            )
            return pa.Table.from_pandas(
                pd.DataFrame(columns=cols), preserve_index=False, schema=schema
            )

    def _process_transform_output(self, output, *args, **kwargs):
        init_num_rows = DataHandler.get_shape(args[0])[0]
        final_num_rows = DataHandler.get_shape(output)[0]
        if final_num_rows != init_num_rows:
            logger.info(
                f'"{self.__class__.__name__}": Resampled to {final_num_rows} from '
                f"{init_num_rows} samples"
            )
        else:
            logger.info(f'"{self.__class__.__name__}": No resampling performed')
        return output

    pass



from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Type

from biocore import DataHandler

from biofit.integration import RCaller
from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.visualization.plotting import BasePlotter, PlotterConfig

if TYPE_CHECKING:
    from datasets import Dataset


@dataclass
class SampleMetadataPlotterConfig(PlotterConfig):
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [
            get_feature("METADATA_FEATURE_TYPES"),
            get_feature("TARGET_FEATURE_TYPES"),
        ],
        init=False,
        repr=False,
    )
    r_source: str = field(
        default=(Path(__file__).parent / "plot_sample_metadata.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="plot_sample_metadata", init=False, repr=False)
    sample_metadata_columns: Optional[SelectedColumnTypes] = None
    outcome_column: Optional[SelectedColumnTypes] = None


class SampleMetadataPlotter(BasePlotter):
    _config_class = SampleMetadataPlotterConfig
    config: SampleMetadataPlotterConfig

    def __init__(
        self,
        config: Optional[SampleMetadataPlotterConfig] = None,
        sample_metadata_columns: Optional[SelectedColumnTypes] = None,
        outcome_column: Optional[SelectedColumnTypes] = None,
        install_missing: bool = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            sample_metadata_columns=sample_metadata_columns,
            outcome_column=outcome_column,
            install_missing=install_missing,
            **kwargs,
        )
        r_source = (Path(__file__).parent / "plot_sample_metadata.R").as_posix()
        r_caller = RCaller.from_script(r_source)
        self.plotter = r_caller.get_method(self.config.main_method)

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        r_source = (Path(__file__).parent / "plot_sample_metadata.R").as_posix()
        r_caller = RCaller.from_script(r_source)
        self.plotter = r_caller.get_method(self.config.main_method)
        return self

    def plot_dataset(self, X: "Dataset", y: "Dataset" = None):
        from biosets import decode

        if y is not None and isinstance(
            next(iter(y._info.features.values())), get_feature("ClassLabel")
        ):
            y = decode(y)
        current_name = next(iter(y._info.features.keys()))
        original_name = next(iter(y._info.features.values())).id or current_name
        if current_name != original_name:
            y = DataHandler.rename_column(y, current_name, original_name)
        if original_name in X.column_names:
            X = DataHandler.drop_column(X, original_name)
        self.plot_arrow(
            DataHandler.to_arrow(X), DataHandler.to_arrow(y) if y is not None else None
        )

    def plot_arrow(self, X, y):
        if self.config.outcome_column is None and y is None:
            raise ValueError(
                "No outcome column provided. Please provide the outcome data "
                "or specify the outcome column."
            )

        if self.config.outcome_column is None:
            self.config.outcome_column = list(y.column_names)[0]
        self.plotter(X, outcome=y, **self.config.get_params())

    def plot(
        self,
        sample_metadata,
        outcome_data: SelectedColumnTypes = None,
        sample_metadata_columns: Optional[SelectedColumnTypes] = None,
        outcome_column: Optional[SelectedColumnTypes] = None,
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(
            sample_metadata_columns, outcome_column
        )
        return self._plot(
            sample_metadata,
            outcome_data,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )



import os
import sys
from typing import TYPE_CHECKING, Optional

import joblib
from biocore.utils.import_util import is_optuna_available

if TYPE_CHECKING:
    import optuna


def report_generation(data, study: Optional["optuna.Study"] = None, cache_dir=None):
    if cache_dir is None and (
        (
            "biosets" in sys.modules
            and isinstance(data, getattr(sys.modules["biosets"], "Bioset"))
        )
        or (
            "datasets" in sys.modules
            and isinstance(data, getattr(sys.modules["datasets"], "Dataset"))
        )
    ):
        cache_dir = os.path.dirname(data.cache_files[0]["filename"])

    if study is None:
        if cache_dir:
            with open(os.path.join(cache_dir, "study.joblib"), "rb") as f:
                study = joblib.load(f)

    if is_optuna_available() and study:
        import optuna.visualization as ov

        if study is None:
            raise ValueError("Study object is required for optuna report generation")
        ov.plot_optimization_history(study)



# ruff: noqa
from .auto import *
from .models import *
from .train import train
from .utils import (
    disable_progress_bar,
    enable_progress_bar,
    logging,
    set_verbosity,
    set_verbosity_debug,
    set_verbosity_error,
    set_verbosity_info,
)
from .utils.version import __version__
from .visualization import *



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Type

import numpy as np

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes
from biofit.utils import logging

from ..stat import Stat, StatConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class ColumnMeanStatConfig(StatConfig):
    # process description
    _transform_process_desc: str = field(
        default="Calculating column means", init=False, repr=False
    )
    processor_name: str = field(default="col_mean", init=False, repr=False)


@dataclass
class ColumnMeanStatConfigForOTU(ColumnMeanStatConfig):
    """Computes the sum of each row in the OTUAbundance feature.
    This class is the same as TotalOTUAbundanceStat.
    It is provided for autostat
    """

    # dataset specific attributes
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    dataset_name: str = field(default="otu", init=False, repr=False)


class ColumnMeanStat(Stat):
    # config attributes
    _config_class = ColumnMeanStatConfig
    config: ColumnMeanStatConfig

    sums_ = None
    counts_ = 0

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "ColumnMeanStat":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_numpy(self, X: np.ndarray, y=None):
        self.config.means = np.mean(X, axis=0)
        return self

    def _partial_fit_numpy(self, X: np.ndarray, y=None):
        self.counts_ += X.shape[0]
        if self.sums_ is None:
            self.sums_ = np.sum(X, axis=0)
        else:
            self.sums_ += np.sum(X, axis=0)

        # self.config.means = self.config.sums_ / self.config.counts_
        return self

    def _fit_polars(self, X: "pl.DataFrame"):
        self.config.means = X.mean()
        return self

    def _partial_fit_polars(self, X: "pl.DataFrame"):
        self.counts_ += X.shape[0]
        if self.sums_ is None:
            self.sums_ = X.sum()
        else:
            self.sums_ += X.sum()

        # self.config.means = self.sums_ / self.counts_
        return self

    def _pool_fit(self, fitted_processors: List["ColumnMeanStat"]):
        self.sums_ = sum([p.sums_ for p in fitted_processors])
        self.counts_ = sum([p.counts_ for p in fitted_processors])
        self.config.means = self.sums_ / self.counts_
        return self

    def _process_transform_output(self, output, input, *args, **kwargs):
        return super()._process_transform_output(
            self.config.means, input, *args, **kwargs
        )



class Unset:
    """A class to represent an unset value.

    This is used to differentiate between a value that is not set and a value that is set to None.
    Optionally, a description can be provided to indicate what the actual default is when Unset is used.
    """

    def __init__(self, description: str = ""):
        self.description = description

    def __repr__(self) -> str:
        """Return the string representation of the class, including any default description if provided.

        Returns:
            The string representation of the class.
        """
        if self.description:
            return f"Unset (default: {self.description})"
        return "Unset"

    def __bool__(self) -> bool:
        """Return False when the class is used in a boolean context.

        Returns:
            False
        """
        return False



"""
Modified by: Patrick Smyth
date: 2024
Summary of changes:

Removed the need of huggingface_hub

datasets
~~~~~~~~
Utilities for working with the local dataset cache.
This file is adapted from the AllenNLP library at https://github.com/allenai/allennlp
Copyright by the AllenNLP authors.
"""

import functools
import io
import os
import posixpath
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Optional, TypeVar, Union
from urllib.parse import unquote, urlparse
import hashlib

from . import logging

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name

INCOMPLETE_SUFFIX = ".incomplete"

PathLike = Union[str, Path]
T = TypeVar("T", str, Path)

_kwargs = {"usedforsecurity": False} if sys.version_info >= (3, 9) else {}
sha256 = functools.partial(hashlib.sha256, **_kwargs)


def is_remote_url(url_or_filename: Union[str, Path]) -> bool:
    return urlparse(str(url_or_filename)).scheme != "" and not os.path.ismount(
        urlparse(str(url_or_filename)).scheme + ":/"
    )


def is_local_path(url_or_filename: Union[str, Path]) -> bool:
    # On unix the scheme of a local path is empty (for both absolute and relative),
    # while on windows the scheme is the drive name (ex: "c") for absolute paths.
    # for details on the windows behavior, see https://bugs.python.org/issue42215
    url_or_filename = Path(url_or_filename).resolve().as_posix()

    return urlparse(url_or_filename).scheme == "" or os.path.ismount(
        urlparse(url_or_filename).scheme + ":/"
    )


def is_relative_path(url_or_filename: Union[str, Path]) -> bool:
    return urlparse(str(url_or_filename)).scheme == "" and not os.path.isabs(
        str(url_or_filename)
    )


def expand_path(path):
    """
    Check if a path is relative and expand it if necessary.
    Handles file paths and URLs, including user home directory expansion.

    :param path: str representing the path or URL to check and expand
    :return: str with the expanded absolute path or full URL
    """
    # Parse the path as a URL
    parsed = urlparse(path)

    # If it's a URL (not a local file path)
    if parsed.scheme and parsed.netloc:
        return path  # Return the original URL

    # If it's a file URL, convert it to a local path
    if parsed.scheme == "file":
        path = unquote(parsed.path)
        # On Windows, remove leading slash
        if sys.platform == "win32" and path.startswith("/"):
            path = path[1:]
    else:
        # It's a regular path, use the full original string
        path = unquote(path)

    # Convert to Path object
    path = Path(path)

    # Expand user's home directory if present
    path = path.expanduser()

    # Check if the path is absolute
    if path.is_absolute():
        return str(path.resolve())

    # If path is relative, make it absolute
    return os.path.normpath(str((Path.cwd() / path).resolve()))


def relative_to_absolute_path(path: T) -> T:
    """Convert relative path to absolute path."""
    abs_path_str = os.path.abspath(os.path.expanduser(os.path.expandvars(str(path))))
    return Path(abs_path_str) if isinstance(path, Path) else abs_path_str


def is_file_name(url_or_path_or_file: T) -> bool:
    if is_local_path(url_or_path_or_file):
        if is_relative_path(url_or_path_or_file):
            if "/" not in Path(url_or_path_or_file).as_posix():
                return True
    return False


def has_ext(url_or_path: Union[str, Path], ext: Optional[str] = None) -> bool:
    if ext is None:
        return Path(url_or_path).suffix != ""
    return Path(url_or_path).suffix == ext


def has_separator(url_or_path: Union[str, Path]) -> bool:
    return "/" in str(url_or_path) or "\\" in str(url_or_path)


def url_or_path_join(base_name: str, *pathnames: str) -> str:
    if is_remote_url(base_name):
        return posixpath.join(
            base_name,
            *(str(pathname).replace(os.sep, "/").lstrip("/") for pathname in pathnames),
        )
    else:
        return Path(base_name, *pathnames).as_posix()


def url_or_path_parent(url_or_path: str) -> str:
    if is_remote_url(url_or_path):
        return url_or_path[: url_or_path.rindex("/")]
    else:
        return os.path.dirname(url_or_path)


def hash_url_to_filename(url, etag=None):
    """
    Convert `url` into a hashed filename in a repeatable way.
    If `etag` is specified, append its hash to the url's, delimited
    by a period.
    If the url ends with .h5 (Keras HDF5 weights) adds '.h5' to the name
    so that TF 2.0 can identify it as a HDF5 file
    (see https://github.com/tensorflow/tensorflow/blob/00fad90125b18b80fe054de1055770cfb8fe4ba3/tensorflow/python/keras/engine/network.py#L1380)
    """
    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    if url.endswith(".py"):
        filename += ".py"

    return filename


def add_start_docstrings(*docstr):
    def docstring_decorator(fn):
        fn.__doc__ = (
            "".join(docstr) + "\n\n" + (fn.__doc__ if fn.__doc__ is not None else "")
        )
        return fn

    return docstring_decorator


def add_end_docstrings(*docstr):
    def docstring_decorator(fn):
        fn.__doc__ = (
            (fn.__doc__ if fn.__doc__ is not None else "") + "\n\n" + "".join(docstr)
        )
        return fn

    return docstring_decorator


def estimate_dataset_size(paths):
    return sum(path.stat().st_size for path in paths)


def readline(f: io.RawIOBase):
    # From: https://github.com/python/cpython/blob/d27e2f4d118e7a9909b6a3e5da06c5ff95806a85/Lib/_pyio.py#L525
    res = bytearray()
    while True:
        b = f.read(1)
        if not b:
            break
        res += b
        if res.endswith(b"\n"):
            break
    return bytes(res)


def move_temp_file(
    temp_file: Union[tempfile._TemporaryFileWrapper, str, Path], final_file: str
):
    if isinstance(temp_file, tempfile._TemporaryFileWrapper):
        temp_file.close()
        temp_file_name = Path(temp_file.name)
    elif not isinstance(temp_file, Path):
        temp_file_name = Path(temp_file)

    if not isinstance(final_file, Path):
        final_file = Path(final_file)

    if temp_file_name.exists():
        if final_file.exists():
            final_file.unlink()
        # is source a windows path?
        if temp_file_name.resolve().drive:
            if len(str(temp_file_name)) > 255:
                src_file = "\\\\?\\" + str(temp_file_name.resolve())
            else:
                src_file = temp_file_name.as_posix()
        else:
            src_file = temp_file_name.as_posix()

        if final_file.resolve().drive:
            if len(str(final_file)) > 255:
                dst_file = "\\\\?\\" + str(final_file.resolve())
            else:
                dst_file = final_file.as_posix()
        else:
            dst_file = final_file.as_posix()
        shutil.move(src_file, dst_file)
        umask = os.umask(0o666)
        os.umask(umask)
        os.chmod(dst_file, 0o666 & ~umask)
        return dst_file
    else:
        raise FileNotFoundError(f"Temporary file {temp_file_name.name} not found.")



import textwrap
from dataclasses import dataclass, field
from typing import List, Optional, Type

import biofit.config as config
from biofit.integration.biosets import get_feature
from biofit.integration.R.r_caller import RCaller
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils.types import Unset
from biofit.visualization.plotting import BasePlotter, PlotterConfig


@dataclass
class ScatterPlotConfig(PlotterConfig):
    processor_type: str = field(default="scaling", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [None, get_feature("TARGET_FEATURE_TYPES"), None],
        init=False,
        repr=False,
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None, None],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES"), None, None],
        init=False,
        repr=False,
    )
    r_source: str = field(
        default=(config.R_SCRIPTS / "plotting_utils.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="generate_scatterplot", init=False, repr=False)

    groupby: str = None
    xdata: str = None
    ydata: str = None
    xlab: str = None
    ylab: str = None
    title: str = "Scatterplot"
    alpha: str = 1
    col_set: str = "Set1"
    cols: List[str] = None
    xlog: str = None
    ylog: str = None


class ScatterPlotter(BasePlotter):
    _config_class = ScatterPlotConfig
    config: ScatterPlotConfig

    def __init__(
        self,
        groupby: str = None,
        xdata: str = None,
        ydata: str = None,
        xlab: str = Unset("None"),
        ylab: str = Unset("None"),
        title: str = Unset('"Scatterplot"'),
        alpha: str = Unset("1"),
        col_set: str = Unset('"Set1"'),
        cols: List[str] = Unset("None"),
        xlog: str = Unset("None"),
        ylog: str = Unset("None"),
        install_missing: bool = None,
        config: Optional[ScatterPlotConfig] = None,
    ):
        super().__init__(
            config=config,
            xlab=xlab,
            ylab=ylab,
            title=title,
            alpha=alpha,
            col_set=col_set,
            cols=cols,
            xlog=xlog,
            ylog=ylog,
            install_missing=install_missing,
        )
        self.plotter = None
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        if self.config.r_source and self.config.main_method:
            self.r_caller = RCaller.from_script(self.config.r_source)
            enter_code = textwrap.dedent(
                """
                suppressPackageStartupMessages(require(ggplot2))
                """
            )
            exit_code = textwrap.dedent(
                """
                ggplot2::ggsave(path, results)
                """
            )
            self.plotter = self.r_caller.get_method(
                self.config.main_method, enter_code, exit_code
            )

        return self

    def plot(
        self,
        x,
        y=None,
        group=None,
        xdata: SelectedColumnTypes = None,
        ydata: SelectedColumnTypes = None,
        groupby: SelectedColumnTypes = None,
        xlab: str = Unset("None"),
        ylab: str = Unset("None"),
        title: str = Unset('"Scatterplot"'),
        alpha: str = Unset("1"),
        col_set: str = Unset('"Set1"'),
        cols: List[str] = Unset("None"),
        xlog: str = Unset("None"),
        ylog: str = Unset("None"),
        path: str = None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        self.config._input_columns = self._set_input_columns_and_arity(
            xdata, ydata, groupby
        )
        return self._plot(
            x,
            y,
            group,
            xlab=xlab,
            ylab=ylab,
            title=title,
            alpha=alpha,
            col_set=col_set,
            cols=cols,
            xlog=xlog,
            ylog=ylog,
            path=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )

    def plot_dataset(self, x, y=None, group=None):
        from biosets import Bioset, decode

        if isinstance(group, Bioset):
            group = decode(group)

        return self.plot_arrow(
            x._data.table,
            y._data.table if y else None,
            group._data.table if group else None,
        )

    def plot_arrow(self, x, y=None, group=None):
        kwargs = self.config.get_params()
        context_kwargs = {
            "path": kwargs.pop("path", None),
        }
        self.plotter(x, y, group, context_kwargs=context_kwargs, **kwargs)



# ruff: noqa
from .pcoa import (
    PCoAFeatureExtractor,
    PCoAFeatureExtractorConfig,
    PCoAFeatureExtractorConfigForOTU,
)
from .plot_pcoa import (
    PCoAFeatureExtractorPlotter,
    PCoAFeatureExtractorPlotterConfig,
)



import os

from biofit.integration.R.r_caller import (
    R_PLOTTING_DEPENDENCIES,
    R_PREPROCESSING_DEPENDENCIES,
    RCaller,
)
from biofit.utils import logging

logger = logging.get_logger(__name__)


def add_to_parser(parser):
    # Register the 'install' subcommand
    install_r_parser = parser.add_parser(
        "install",
        help="Install R dependencies",
        description="Install various dependencies including R and its packages.",
    )

    install_r_parser.add_argument(
        "--all",
        action="store_true",
        help="Install all R dependencies for plotting and preprocessing",
    )
    install_r_parser.add_argument(
        "--plotting",
        action="store_true",
        help="Install R plotting dependencies",
    )
    install_r_parser.add_argument(
        "--preprocessing",
        action="store_true",
        help="Install R preprocessing dependencies",
    )
    install_r_parser.add_argument(
        "--cran",
        nargs="+",
        help="Install specific CRAN packages",
    )
    install_r_parser.add_argument(
        "--bioconductor",
        nargs="+",
        help="Install specific Bioconductor packages",
    )
    install_r_parser.add_argument(
        "--binary",
        action="store_true",
        help="Install with binaries only",
    )

    install_r_parser.add_argument(
        "--r-home",
        "-r",
        type=str,
        help="Path to the R installation directory which contains the library folder",
    )
    return parser


def run(args):
    if args.r_home:
        os.environ["R_HOME"] = args.r_home
    cran_deps = args.cran or []
    bioconductor_deps = args.bioconductor or []
    params = {}
    if args.binary:
        params["pkgType"] = "binary"
    if args.all or args.plotting:
        cran_deps += R_PLOTTING_DEPENDENCIES["cran"]
        bioconductor_deps += R_PLOTTING_DEPENDENCIES["bioconductor"]
    if args.all or args.preprocessing:
        cran_deps += R_PREPROCESSING_DEPENDENCIES["cran"]
        bioconductor_deps += R_PREPROCESSING_DEPENDENCIES["bioconductor"]

    if cran_deps:
        logger.info(f"Checking for CRAN packages: {', '.join(cran_deps)}")
        RCaller.verify_r_dependencies(
            cran_dependencies=cran_deps,
            bioconductor_dependencies=[],
            install_missing=True,
        )
        logger.info("CRAN dependencies installed successfully.")

    if bioconductor_deps:
        logger.info(f"Checking Bioconductor packages: {', '.join(bioconductor_deps)}")
        RCaller.verify_r_dependencies(
            cran_dependencies=[],
            bioconductor_dependencies=bioconductor_deps,
            install_missing=True,
        )
        logger.info("Bioconductor dependencies installed successfully.")

    if not cran_deps and not bioconductor_deps:
        logger.info("No R dependencies to install.")



"""
Feature selector that filters features based on the number of samples they are present in.
"""

from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Type

import numpy as np
from biocore import DataHandler

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.stat import ColumnMissingnessStat, ColumnMissingnessStatConfig
from biofit.utils import logging

from ..feature_selection import FeatureSelector, FeatureSelectorConfig

logger = logging.get_logger(__name__)

FILTER_FEATURES_DOCSTRING = """
SampleFilter features based on the number of samples they are present in.

Args:
    X: Input data
    min_prevalence: Minimum number of samples a feature must be present in to be kept.

Returns:
    SampleFiltered data.
"""


def _filter_features(nrows: int, total_missing, min_prevalence, cols=None):
    """
    SampleFilter features in a pandas DataFrame based on their presence in the dataset.

    Args:
        X (pd.DataFrame): The input DataFrame containing the features.
        min_prevalence (float, optional): The minimum required presence of a feature in the dataset.
        depth (float, optional): The minimum value to be considered as present.

    Returns:
        List[int]: A list of column indices that pass the filtering condition.
    """
    # keep features that are present in at least min_prevalence of the rows
    col_to_keep = total_missing <= (nrows * (1 - min_prevalence))
    if cols and len(cols) == len(col_to_keep):
        return [c for c, b in zip(cols, col_to_keep) if b]
    else:
        return [i for i, b in enumerate(col_to_keep) if b]


@dataclass
class MinPrevalenceFeatureSelectorConfig(FeatureSelectorConfig):
    processor_name: str = field(
        default="min_prevalence_feature_selector", init=False, repr=False
    )
    _fit_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )

    min_prevalence: float = 0.5
    depth: int = None

    def __post_init__(self):
        if self.depth is None:
            if isinstance(self.min_prevalence, float) and self.min_prevalence < 1:
                self._fit_process_desc = (
                    "Removing features that are present in less than "
                    f"{self.min_prevalence * 100:.0f}% of samples"
                )
            elif (
                isinstance(self.min_prevalence, (int, float))
                and self.min_prevalence >= 1
            ):
                self._fit_process_desc = f"Removing features that are present in less than {self.min_prevalence} samples"

        elif isinstance(self.min_prevalence, float) and self.min_prevalence < 1:
            self._fit_process_desc = (
                f"Removing features with <{self.min_prevalence * 100:.0f}% "
                f"samples above {self.depth} counts"
            )
        elif isinstance(self.min_prevalence, (int, float)) and self.min_prevalence >= 1:
            self._fit_process_desc = (
                f"Removing features with <{self.min_prevalence} "
                f"samples above {self.depth} counts"
            )


@dataclass
class MinPrevalenceFeatureSelectorConfigForMetagenomics(
    MinPrevalenceFeatureSelectorConfig
):
    dataset_name: str = field(default="metagenomics", init=False, repr=False)
    _fit_input_feature_types: List[Tuple[Type, Type]] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    _transform_input_feature_types: List[Tuple[Type, Type]] = field(
        default_factory=lambda: [(get_feature("Abundance"), get_feature("ReadCount"))],
        init=False,
        repr=False,
    )
    depth: int = 100
    min_prevalence: float = 0.01


@dataclass
class MinPrevalenceFeatureSelectorConfigForOTU(MinPrevalenceFeatureSelectorConfig):
    dataset_name: str = field(default="otu", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("Abundance")], init=False, repr=False
    )
    depth: int = 0
    min_prevalence: float = 0.01

    def __post_init__(self):
        if self.depth is None:
            if isinstance(self.min_prevalence, float) and self.min_prevalence < 1:
                self._fit_process_desc = (
                    "Removing OTUs that are present in less than "
                    f"{self.min_prevalence * 100:.0f}% of samples"
                )
            elif (
                isinstance(self.min_prevalence, (int, float))
                and self.min_prevalence >= 1
            ):
                self._fit_process_desc = f"Removing OTUs that are present in less than {self.min_prevalence} samples"

        elif isinstance(self.min_prevalence, float) and self.min_prevalence < 1:
            self._fit_process_desc = (
                f"Removing OTUs with <{self.min_prevalence * 100:.0f}% "
                f"samples above {self.depth} counts"
            )
        elif isinstance(self.min_prevalence, (int, float)) and self.min_prevalence >= 1:
            self._fit_process_desc = (
                f"Removing OTUs with <{self.min_prevalence} "
                f"samples above {self.depth} counts"
            )


@dataclass
class MinPrevalenceFeatureSelectorConfigForSNP(MinPrevalenceFeatureSelectorConfig):
    dataset_name: str = field(default="snp", init=False, repr=False)
    _fit_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    _transform_input_feature_types: List[Type] = field(
        default_factory=lambda: [get_feature("GenomicVariant")], init=False, repr=False
    )
    depth: int = 0
    min_prevalence: float = 0.012


class MinPrevalenceFeatureSelector(FeatureSelector):
    """
    Feature selector that filters features based on the number of samples they are present in.

    - config:
    - min_prevalence (Union[str, float], optional):
        The minimum prevalence of features to keep a sample.
        The threshold to which we determine the minimum prevalence percentage or count of present values to maintain a sample or feature.
        Any value passed >= 1 will be the minimum count while any value < 1 will be a percentage of the total values.
        If "auto", the minimum prevalence is calculated as the first quartile minus 1.5 times the interquartile range. Defaults to "auto".
    - depth (Union[int, Unset], optional):
        The minimum value that we consider something to be present. Defaults to None.
    """

    config_class = MinPrevalenceFeatureSelectorConfig
    config: MinPrevalenceFeatureSelectorConfig

    def __init__(
        self,
        min_prevalence: float = 0.5,
        depth: int = None,
        config: Optional[MinPrevalenceFeatureSelectorConfig] = None,
        **kwargs,
    ):
        super().__init__(
            config=config, depth=depth, min_prevalence=min_prevalence, **kwargs
        )
        self = self.set_params()

    @sync_backup_config
    def set_params(self, **kwargs):
        if len(kwargs) > 0:
            self.config = self.config.replace_defaults(**kwargs)
        col_missingness_config = ColumnMissingnessStatConfig.from_config(self.config)
        col_missingness_config = col_missingness_config.replace_defaults(**kwargs)
        self.missingness = ColumnMissingnessStat(config=col_missingness_config)
        self.total_missing = None
        self._input_columns = None
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "MinPrevalenceFeatureSelector":
        self.config._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(input_columns)
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        *args,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _process_fit_input(self, input, **kwargs):
        self.height = DataHandler.get_shape(input)[0]
        return input, kwargs

    def _fit_numpy(self, X):
        self.total_missing = self.missingness._transform_numpy(X).flatten()
        return self

    def _partial_fit_numpy(self, X):
        if self.total_missing is None:
            self.total_missing = self.missingness._transform_numpy(X).flatten()
        else:
            self.total_missing += self.missingness._transform_numpy(X).flatten()
        return self

    def _fit_pandas(self, X):
        self.total_missing = self.missingness._transform_pandas(X).values.flatten()
        return self

    def _partial_fit_pandas(self, X):
        if self.total_missing is None:
            self.total_missing = self.missingness._transform_pandas(X).values.flatten()
        else:
            self.total_missing += self.missingness._transform_pandas(X).values.flatten()
        return self

    def _fit_polars(self, X):
        self.total_missing = self.missingness._transform_polars(X).to_numpy().flatten()
        return self

    def _partial_fit_polars(self, X):
        if self.total_missing is None:
            self.total_missing = (
                self.missingness._transform_polars(X).to_numpy().flatten()
            )
        else:
            self.total_missing += (
                self.missingness._transform_polars(X).to_numpy().flatten()
            )
        return self

    def _fit_arrow(self, X):
        self.total_missing = np.array(
            [v[0] for _, v in self.missingness._transform_arrow(X).to_pydict().items()]
        )
        return self

    def _partial_fit_arrow(self, X):
        if self.total_missing is None:
            total_missing = self.missingness._transform_arrow(X)
            self.total_missing = np.array(
                [v[0] for _, v in total_missing.to_pydict().items()]
            )
        else:
            other_missing = self.missingness._transform_arrow(X)
            other_missing = np.array(
                [v[0] for _, v in other_missing.to_pydict().items()]
            )
            self.total_missing += other_missing
        return self

    def _pool_fit(self, out: List["MinPrevalenceFeatureSelector"]):
        new_self = out[0]
        if len(out) > 1:
            logger.info("Pooling results")
            new_self.total_missing = np.sum(
                np.vstack([x.total_missing for x in out]), axis=0
            )
        logger.info(
            f"Total missing values: {new_self.total_missing.sum()} "
            f"({new_self.total_missing.mean():.2f} per feature)"
        )
        return new_self

    def _process_fit_output(self, input, out):
        self.config._feature_idx_out = _filter_features(
            self.height,
            self.total_missing,
            self.config.min_prevalence,
        )
        return super()._process_fit_output(input, out)



from collections import OrderedDict

from biofit.auto.auto_factory import _BaseAutoModelClass, _LazyAutoMapping

from .configuration_auto import CONFIG_MAPPING_NAMES

MODEL_MAPPING_NAMES = OrderedDict(
    [
        ("xgboost", "XGBoostModel"),
        ("lightgbm", "LightGBMModel"),
        ("catboost", "CatBoostModel"),
        ("random_forest", "RandomForestModel"),
        ("logistic_regression", "LogisticRegressionModel"),
        ("lasso", "LassoModel"),
        ("svm", "SVMModel"),
        ("knn", "KNNModel"),
    ]
)

MODEL_FOR_CLASSIFICATION_MAPPING_NAMES = OrderedDict(
    [
        ("xgboost", "XGBoostForClassification"),
        ("lightgbm", "LightGBMForClassification"),
        ("catboost", "CatBoostForClassification"),
        ("random_forest", "RandomForestForClassification"),
        ("logistic_regression", "LogisticRegressionForClassification"),
        ("lasso", "LassoForClassification"),
        ("svm", "SVMForClassification"),
        ("knn", "KNNForClassification"),
    ]
)

MODEL_FOR_REGRESSION_MAPPING_NAMES = OrderedDict(
    [
        ("xgboost", "XGBoostForRegression"),
        ("lightgbm", "LightGBMForRegression"),
        ("catboost", "CatBoostForRegression"),
        ("random_forest", "RandomForestForRegression"),
        ("logistic_regression", "LogisticRegressionForRegression"),
        ("lasso", "LassoForRegression"),
        ("svm", "SVMForRegression"),
        ("knn", "KNNForRegression"),
    ]
)

MODEL_MAPPING = _LazyAutoMapping(CONFIG_MAPPING_NAMES, MODEL_MAPPING_NAMES)
MODEL_FOR_CLASSIFICATION_MAPPING_NAMES = _LazyAutoMapping(
    CONFIG_MAPPING_NAMES, MODEL_FOR_CLASSIFICATION_MAPPING_NAMES
)


class AutoModel(_BaseAutoModelClass):
    _processor_mapping = MODEL_MAPPING


class AutoModelForClassification(_BaseAutoModelClass):
    _processor_mapping = MODEL_FOR_CLASSIFICATION_MAPPING_NAMES


class AutoModelForRegression(_BaseAutoModelClass):
    _processor_mapping = MODEL_FOR_REGRESSION_MAPPING_NAMES



# ruff: noqa
from .css import (
    CumulativeSumScaler,
    CumulativeSumScalerConfig,
    CumulativeSumScalerConfigForOTU,
    CumulativeSumScalerConfigForMetagenomics,
    # CumulativeSumScalerConfigForGenomics,
    # CumulativeSumScalerConfigForSNP,
)
from .plot_css import (
    CumulativeSumScalerPlotter,
    CumulativeSumScalerPlotterConfig,
    CumulativeSumScalerPlotterConfigForMetagenomics,
    CumulativeSumScalerPlotterConfigForOTU,
    CumulativeSumScalerPlotterConfigForASV,
    CumulativeSumScalerPlotterConfigForGenomics,
    CumulativeSumScalerPlotterConfigForSNP,
    CumulativeSumScalerPlotterConfigForReadCount,
)



from dataclasses import dataclass, field
from typing import Optional

from biofit.processing import SelectedColumnTypes
from biofit.utils.types import Unset
from biofit.visualization.plotting_utils import plot_dimension_reduction

from ..plot_feature_extraction import (
    FeatureExtractorPlotter,
    FeatureExtractorPlotterConfig,
)


@dataclass
class PCoAFeatureExtractorPlotterConfig(FeatureExtractorPlotterConfig):
    processor_name: str = field(default="pcoa", init=False, repr=False)
    title: str = "PCoA Plot"
    n_components: int = 3
    label_column: str = None
    group_column: str = None


class PCoAFeatureExtractorPlotter(FeatureExtractorPlotter):
    config_class = PCoAFeatureExtractorPlotterConfig
    config: PCoAFeatureExtractorPlotterConfig

    def __init__(
        self,
        n_components: int = 3,
        label_column: str = None,
        group_column: str = None,
        config: Optional[PCoAFeatureExtractorPlotterConfig] = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            n_components=n_components,
            label_column=label_column,
            group_column=group_column,
            **kwargs,
        )

    def plot(
        self,
        X,
        labels=None,
        group=None,
        input_columns: SelectedColumnTypes = None,
        label_column: SelectedColumnTypes = None,
        group_column: SelectedColumnTypes = None,
        precomputed: bool = Unset("False"),
        pca_kwargs: dict = Unset("{}"),
        title: str = Unset("PCA Plot"),
        n_components: int = Unset(3),
        path=None,
        device: str = "pdf",
        fingerprint: str = None,
        unused_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = True,
        show: bool = True,
    ):
        """Plot the PCoA plot.

        Args:
            X: The input data.
            labels: The labels for the data.
            group: The group for the data.
            input_columns: The input columns.
            label_column: The label column.
            group_column: The group column.
            precomputed: Whether the input data is precomputed.
            pcoa_kwargs: The additional kwargs for computing PCoA. Only used if precomputed is False.
            **kwargs: The additional kwargs.

        """
        return plot_dimension_reduction(
            X,
            labels=labels,
            group=group,
            input_columns=input_columns,
            label_column=label_column,
            group_column=group_column,
            method="pcoa" if not precomputed else None,
            method_kwargs=pca_kwargs,
            output_dir=path,
            device=device,
            fingerprint=fingerprint,
            unused_columns=unused_columns,
            raise_if_missing=raise_if_missing,
            show=show,
        )



from dataclasses import dataclass, field
from pathlib import Path

from biocore import DataHandler
from biocore.utils.import_util import is_biosets_available

from biofit.integration.biosets import get_feature
from biofit.processing import SelectedFeatureTypes
from biofit.visualization.plotting import BasePlotter, PlotterConfig


@dataclass
class ScalerPlotterConfig(PlotterConfig):
    processor_type: str = field(default="scaling", init=False, repr=False)
    r_source: str = field(
        default=(Path(__file__).parent / "plot_scaling.R").as_posix(),
        init=False,
        repr=False,
    )
    main_method: str = field(default="plot_scaler", init=False, repr=False)
    _fit_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )
    _transform_unused_feature_types: SelectedFeatureTypes = field(
        default_factory=lambda: [get_feature("METADATA_FEATURE_TYPES")],
        init=False,
        repr=False,
    )


class ScalerPlotter(BasePlotter):
    config_class = ScalerPlotterConfig
    config: ScalerPlotterConfig

    def plot_dataset(self, x1, x2, y1, y2):
        if is_biosets_available():
            from biosets import decode

            y1 = decode(y1) if y1 is not None else None
            y2 = decode(y2) if y2 is not None else None
        return self.plot_arrow(
            DataHandler.to_arrow(x1),
            DataHandler.to_arrow(x2),
            DataHandler.to_arrow(y1) if y1 is not None else None,
            DataHandler.to_arrow(y2) if y2 is not None else None,
        )

    def plot_arrow(self, x1, x2, y1, y2):
        return self.plotter(x1, x2, y1, y2, **self.config.get_params())



# ruff: noqa
from .row_missingness import (
    RowMissingnessStat,
    RowMissingnessStatConfig,
    RowMissingnessStatConfigForOTU,
    RowMissingnessStatConfigForSNP,
    RowMissingnessStatConfigForMetagenomics,
)
